#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
#
# Network device status
#
# Copyright (c) 2005 JAS
#
# Author: Petr Vokac <vokac@kmlinux.fjfi.cvut.cz>
#
# $Id: arpsniff.py 13 2006-04-27 01:36:19Z vokac $
#
import sys, getopt, time
import logging
import socket
import MySQLdb

__version__ = "$Revision: 13 $"

__DATABASE_HOST = "localhost"
__DATABASE_DB = "dbname"
__DATABASE_USER = "dbuser"
__DATABASE_PASS = "dbpass"

__DATABASE_TABLE = {
    'arp_history': """
    CREATE TABLE IF NOT EXISTS `arp_history` (
    `ip` INT UNSIGNED NOT NULL,
    `mac` CHAR( 17 ) DEFAULT '00:00:00:00:00:00' NOT NULL,
    `changed` DATETIME NOT NULL,
    PRIMARY KEY ( `ip`, `changed` ),
    KEY ( `mac` )
    );
    """,
    'arp_last': """
    CREATE TABLE IF NOT EXISTS `arp_last` (
    `ip` INT UNSIGNED NOT NULL,
    `mac` CHAR( 17 ) DEFAULT '00:00:00:00:00:00' NOT NULL,
    `changed` DATETIME NOT NULL,
    `updated` DATETIME NOT NULL,
    PRIMARY KEY ( `ip` ),
    KEY ( `mac` )
    );
    """,
    }

def getConnection():
    conn = None
    try:
        conn = MySQLdb.connect(host = __DATABASE_HOST,
                               db = __DATABASE_DB,
                               user = __DATABASE_USER,
                               passwd = __DATABASE_PASS)
        cursor = conn.cursor()
        for sql in __DATABASE_TABLE.values():
            logging.getLogger().debug("SQL: %s" % sql)
            cursor.execute(sql)
        cursor.close()
    except MySQLdb.DatabaseError, err:
        logging.getLogger().error("database error: %s" % err)
        raise err
    return conn

def ipToInt(ip):
    retVal = 0
    for x in socket.inet_aton(ip):
        retVal <<= 8
        retVal += ord(x)
    return retVal

def intToIp(i):
    retVal = []
    for x in [ 24, 16, 8, 0 ]:
        retVal.append(str((i >> x) & 0xFF))
    return ".".join(retVal)

def getLastStatus():
    retVal = {}
    conn = getConnection()
    cursor = conn.cursor()
    sql = "SELECT `ip`, LOWER(`mac`) AS `mac`, UNIX_TIMESTAMP() - UNIX_TIMESTAMP(`changed`) as `changed`, UNIX_TIMESTAMP() - UNIX_TIMESTAMP(`updated`) as `updated` FROM `arp_last`"
    logging.getLogger().debug("SQL: %s" % sql)
    cursor.execute(sql)
    while (True):
        row = cursor.fetchone()
        if row == None: break
        retVal[intToIp(row[0])] = (row[1], int(row[2]), int(row[3]))
    cursor.close()
    return retVal

def readLines(ssl):
    lines = []
    while 1:
        data = ssl.read(1024).split("\n")
        if len(lines) == 0:
            lines += data
        else:
            lines[len(lines)-1] += data[0]
            lines += data[1:]
        if len(lines) >= 2 and lines[len(lines)-2] == '' and lines[len(lines)-1] == '':
            break
    lines.pop()
    lines.pop()
    return lines

def updateIpStatus(ip, cursor, status, status_last):
    logging.getLogger().debug("IP: %s" % ip)
    iip = ipToInt(ip)

    if not status_last.has_key(ip):
        if not status.has_key(ip):
            return
        if status[ip][0] == '00:00:00:00:00:00':
            return
                    
    insert_history = False
    (mac, first, last, update) = status[ip]

    # don't process items that was not scanned in last 3 minutes
    if int(update) >= 60*3:
        status[ip] = None
            
    if status_last.has_key(ip):
        (mac_last, changed_last, updated_last) = status_last[ip]
        # records older then 10 minutes are considered to be dead
        if last >= 60*10:
            first = last
            mac = '00:00:00:00:00:00'
        if mac != mac_last:
            insert_history = True
        sql = "UPDATE `arp_last` SET `mac` = '%s', `changed` = FROM_UNIXTIME(UNIX_TIMESTAMP()-%i), `updated` = NOW() WHERE `ip` = %i" % (mac, last, iip)
        logging.getLogger().debug("SQL: %s" % sql)
        cursor.execute(sql)
    else:
        insert_history = True
        sql = "INSERT INTO `arp_last` (`ip`, `mac`, `changed`, `updated`) VALUES (%i, '%s', FROM_UNIXTIME(UNIX_TIMESTAMP()-%i), NOW())" % (iip, mac, last)
        logging.getLogger().debug("SQL: %s" % sql)
        cursor.execute(sql)
    if insert_history:
        sql = "INSERT INTO `arp_history` (`ip`, `mac`, `changed`) VALUES (%i, '%s', FROM_UNIXTIME(UNIX_TIMESTAMP()-%i))" % (iip, mac, first)
        logging.getLogger().debug("SQL: %s" % sql)
        cursor.execute(sql)


def main(host, port, nets = [], scan = False, user = '', password = ''):
    logging.getLogger().debug("main('%s', %s, %s, %s, '%s', '%s')" % (host, port, nets, scan, user, password))

    status_last = getLastStatus()

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((host, port))
    ssl = socket.ssl(s, None, None)

    logging.getLogger().debug("SOCKET: auth")
    ssl.write("auth %s %s\n" % (user, password))
    lines = readLines(ssl)
    if lines[len(lines)-1][:4] == 'fail':
        logging.getLogger().error("auth failed")
        sys.exit(1)

    # scan
    if scan:
        if nets == []:
            logging.getLogger().debug("SOCKET: scan")
            ssl.write("scan\n")
            readLines(ssl)
        else:
            for ip_from, ip_to in nets:
                logging.getLogger().debug("SOCKET: scan %s %s" % (ip_from, ip_to))
                ssl.write("scan %s %s\n" % (ip_from, ip_to))
                readLines(ssl)
                time.sleep(0.1) # don't flood network with ARP requests too fast
        time.sleep(0.5) # give us time to process ARP responses

    logging.getLogger().debug("SOCKET: dump")
    ssl.write("dump\n")
    lines = readLines(ssl)
    status = {}
    for line in lines:
        (ip, mac, first, last, update) = line.split(" ")
        status[ip] = (mac.lower(), int(first), int(last), int(update))

    conn = getConnection()
    cursor = conn.cursor()

    if nets == []:
        for ip in status.keys():
            updateIpStatus(ip, cursor, status, status_last)
    else:
        for ip_from, ip_to in nets:
            for iip in range(ipToInt(ip_from), ipToInt(ip_to)):
                ip = intToIp(iip)
        
                if scan and not status.has_key(ip):
                    status[ip] = ('00:00:00:00:00:00', 0, 0, 0)

                updateIpStatus(ip, cursor, status, status_last)

    cursor.close()

    logging.getLogger().debug("SOCKET: quit")
    ssl.write("quit\n")
    s.close()


def usage():
    print "usage: %s [-s host] [-p port] [network/mask [network/mask [...]]]" % sys.argv[0]
    print "  -v      verbose"
    print "  -h      this help"
    print "  -s host hostname with running arpsniff daemon"
    print "  -p port listening port of arpsniff daemon"
    print "  -u user username for authentization"
    print "  -a pass password for authentization"


if __name__ == '__main__':
    streamHandler = logging.StreamHandler()
    streamHandler.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s](%(module)s:%(lineno)d) %(message)s", "%d %b %H:%M:%S"))
    logging.getLogger().addHandler(streamHandler)
    logging.getLogger().setLevel(logging.DEBUG)

    try:
        opts, args = getopt.getopt(sys.argv[1:], "hvqs:p:u:a:c",
                                   ["help", "verbose", "quiet", "station", "port",
                                    "user", "pass", "scan" ])
    except getopt.GetoptError:
        usage()
        sys.exit(2)
    station = None
    port = 9320
    user = ''
    password = ''
    scan = False
    nets = []
    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit()
        if o in ("-v", "--verbose"):
            logging.getLogger().setLevel(logging.DEBUG)
        if o in ("-q", "--quiet"):
            logging.getLogger().setLevel(logging.ERROR)
        if o in ("-s", "--station"):
            station = a
        if o in ("-p", "--port"):
            port = int(a)
        if o in ("-u", "--user"):
            user = a
        if o in ("-a", "--pass"):
            password = a
        if o in ("-c", "--scan"):
            scan = True
    if len(args) > 0:
        for net in args:
            if net.find("/") != -1:
                net, mask = net.split("/")
                mask = int(mask)
                ip_from = intToIp(ipToInt(net) & ~(0xffffffff >> mask))
                ip_to = intToIp((ipToInt(net) & ~(0xffffffff >> mask)) + (1 << (32 - mask)) - 1)
            else:
                ip_from = net
                ip_to = net
            nets.append((ip_from, ip_to))
            logging.getLogger().debug("nets: %s %s" % (ip_from, ip_to))

    if station == None:
        usage()
        sys.exit()
    else:
        main(station, port, nets, scan, user, password)
